import binascii
import hashlib

from datetime import datetime
from flask import Response, abort, request
from io import BytesIO
from werkzeug import parse_date, http_date


class RangeRequest:

    def __init__(self,
                 data,
                 etag: str=None,
                 last_modified: datetime=None,
                 size=None) -> None:

        if not ((etag is None and last_modified is None and size is None) or
                (etag is not None and last_modified is not None and size is not None)):
            raise ValueError('Must specifiy all range data or none.')

        if isinstance(data, bytes):
            self.__data = BytesIO(data)
        elif isinstance(data, str):
            self.__data = BytesIO(data.encode('utf-8'))
        else:
            self.__data = data

        if etag is not None:
            self.__etag = etag
        else:
            self.__etag = self.make_etag(self.__data)
            self.__data.seek(0)

        if last_modified is not None:
            self.__last_modified = last_modified
        else:
            self.__last_modified = datetime.utcnow()

        if size is not None:
            self.__size = size
        else:
            # TODO this is stupid, but good enough for now
            self.__size = 0
            while True:
                chunk = self.__data.read(4096)
                if chunk:
                    self.__size += len(chunk)
                else:
                    break
            self.__data.seek(0)

    def make_response(self) -> Response:
        use_default_range = True
        status_code = 200
        # range requests are only allowed for get
        if request.method == 'GET':
            range_header = request.headers.get('Range')

            ranges = self.parse_range_header(range_header, self.__size)
            if not (len(ranges) == 1 and ranges[0][0] == 0 and ranges[0][1] == self.__size - 1):
                use_default_range = False
                status_code = 206

            if range_header:
                if_range = request.headers.get('If-Range')
                if if_range and if_range != self.__etag:
                    use_default_range = True
                    status_code = 200

        if use_default_range:
            ranges = [(0, self.__size - 1)]

        if len(ranges) > 1:
            abort(416)  # We don't support multipart range requests yet

        if_unmod = request.headers.get('If-Unmodified-Since')
        if if_unmod:
            if_date = parse_date(if_unmod)
            if if_date and if_date < self.__last_modified:
                status_code = 304

        # TODO If-None-Match support

        if status_code != 304:
            resp = Response(self.generate(ranges, self.__data))
        else:
            resp = Response()

        resp.headers['Content-Length'] = ranges[0][1] - ranges[0][0]
        resp.headers['Accept-Ranges'] = 'bytes'
        resp.headers['ETag'] = self.__etag
        resp.headers['Last-Modified'] = http_date(self.__last_modified)

        if status_code == 206:
            resp.headers['Content-Range'] = \
                'bytes {}-{}/{}'.format(ranges[0][0], ranges[0][1], self.__size)

        resp.status_code = status_code

        return resp

    def generate(self, ranges: list, readable):
        for (start, end) in ranges:
            readable.seek(start)
            bytes_left = end - start + 1

            chunk_size = 4096
            while bytes_left > 0:
                read_size = min(chunk_size, bytes_left)
                chunk = readable.read(read_size)
                bytes_left -= read_size
                yield chunk

    @classmethod
    def parse_range_header(cls, range_header: str, target_size: int) -> list:
        end_index = target_size - 1
        if range_header is None:
            return [(0, end_index)]

        bytes_ = 'bytes='
        if not range_header.startswith(bytes_):
            abort(416)

        ranges = []
        for range_ in range_header[len(bytes_):].split(','):
            split = range_.split('-')
            if len(split) == 1:
                try:
                    start = int(split[0])
                    end = end_index
                except ValueError:
                    abort(416)
            elif len(split) == 2:
                start, end = split[0], split[1]
                if not start:
                    # parse ranges of the form "bytes=-100" (i.e., last 100 bytes)
                    end = end_index
                    try:
                        start = end - int(split[1]) + 1
                    except ValueError:
                        abort(416)
                else:
                    # parse ranges of the form "bytes=100-200"
                    try:
                        start = int(start)
                        if not end:
                            end = target_size
                        else:
                            end = int(end)
                    except ValueError:
                        abort(416)

                    if end < start:
                        abort(416)

                    end = min(end, end_index)
            else:
                abort(416)

            ranges.append((start, end))

        # merge the ranges
        merged = []
        ranges = sorted(ranges, key=lambda x: x[0])
        for range_ in ranges:
            # initial case
            if not merged:
                merged.append(range_)
            else:
                # merge ranges that are adjacent or overlapping
                if range_[0] <= merged[-1][1] + 1:
                    merged[-1] = (merged[-1][0], max(range_[1], merged[-1][1]))
                else:
                    merged.append(range_)

        return merged

    @classmethod
    def make_etag(cls, data):
        hasher = hashlib.sha256()

        while True:
            read_bytes = data.read(4096)
            if read_bytes:
                hasher.update(read_bytes)
            else:
                break

        hash_value = binascii.hexlify(hasher.digest()).decode('utf-8')
        return '"sha256:{}"'.format(hash_value)
